from xl_router.router import Router
from flask import Flask, request 
from flask.json import JSONEncoder
from importlib import import_module
import decimal


class JsonEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, decimal.Decimal):
            return float(obj)
        return JSONEncoder.default(self, obj)


class App(Flask):
    resources = []

    def __init__(self, config, *args, **kwargs):
        super().__init__('', *args, **kwargs)
        self.json_provider_class = JsonEncoder
        self.json.ensure_ascii = False
        if isinstance(config, dict):
            self.config.from_mapping(config)
        elif '.py' in config:
            self.config.from_pyfile(config)

    def register_extensions(self):
        pass

    def register_resources(self):
        for module_name in self.resources:
            module = import_module('app.{}.resources'.format(module_name))
            self.register_blueprint(module.router)


def get_user_agent():
    return request.user_agent.string.lower()


def get_ip():
    nodes = request.headers.getlist("X-Forwarded-For")
    return nodes[0] if nodes else request.remote_addr


def get_rule():
    return request.url_rule


def get_platform():
    user_agent = get_user_agent()
    if 'windows' in user_agent:
        return 1
    if 'mac os' in user_agent and 'iphone' not in user_agent:
        return 1
    return 2


from zmemory import Memory
from flask import current_app


TOKEN_EXPIRATION_SECONDS = 1800


_memory_instance = None


def get_instance():
    global _memory_instance
    if _memory_instance is None:
        _memory_instance = Memory(current_app.REDIS_CONFIG)
    return _memory_instance


def get_cache():
    return Memory({**current_app.REDIS_CONFIG, 'db': 15})


def set_token(token, user_info=None):
    memory = get_instance()
    token_key = f'access_token_{token}'
    memory.set(token_key, user_info, seconds=TOKEN_EXPIRATION_SECONDS)





import re
from typing import Callable


def lowcase(text: str) -> str:
    """驼峰转下划线
    lowcase('sayHello')
    """
    ls = list(set(re.findall('[a-z][A-Z]', text)))
    ll = [x[0] + '_' + chr(ord(x[1]) + 32) for x in ls]
    for x, y in zip(ls, ll):
        text = text.replace(x, y)
    return text

def camel(text: str) -> str:
    """下划线转驼峰"""
    if isinstance(text, str):
        if '_' in text:
            first, *others = text.split('_')
            return ''.join([first.lower(), *map(str.title, others)])
        else:
            return text
    else:
        return text

def to_lowcase(data: dict):
    """将字典键转化为小写"""
    return {lowcase(k): data.get(k) for k in data}

def to_camel(data: dict):
    """"将字典键转化为驼峰"""
    return {camel(k): data.get(k) for k in data}

def iter_list(data: list, func: Callable):
    """深层遍历字典数组，将键进行转化"""
    rt_list = []
    for item in data:
        if isinstance(item, dict):
            rt_list.append(iter_dict(item, func))
        elif isinstance(item, list):
            rt_list.append(iter_list(item, func))
        else:
            rt_list.append(item)
    return rt_list

def iter_dict(data: dict, func: Callable) -> dict:
    """深层遍历字典，将键进行转化"""
    rt_dict = {}
    for k, v in data.items():
        if isinstance(v, dict):
            rt_dict[func(k)] = iter_dict(v, func)
        elif isinstance(v, list):
            rt_dict[func(k)] = iter_list(v, func)
        else:
            rt_dict[func(k)] = v
    return rt_dict

def iter_lowcase(data: dict) -> dict:
    """深层遍历字典，将键转化为小写"""
    return iter_dict(data, lowcase)

def iter_camel(data: dict) -> dict:
    """深层遍历字典，将键转化为驼峰"""
    return iter_dict(data, camel)


from typing import Optional


class R:
    @classmethod
    def get(cls, id: Optional[int] = None, order_key: Optional[str]=None, order_way: Optional[str]=None, page_num: Optional[int] = None, page_size: Optional[int] = None, **kwargs):
        if id:
            return cls.get_json(id)
        else:
            return cls.get_jsons(order_key=order_key, order_way=order_way, page_num=page_num, page_size=page_size)

    @classmethod
    def post(cls, data: dict):
        cls.add(data)

    @classmethod
    def put(cls, id: int, data: dict):
        cls.save(id, data)

    @classmethod
    def delete(cls, id: int):
        cls.delete_list(id=id)

   


from xl_router.exceptions import MessagePrompt
from xl_router.utils.base.json import to_lowcase, iter_lowcase, iter_camel
from flask import Blueprint, request
from jsonschema.exceptions import ValidationError
import functools
import random
import os
import importlib
import inspect
import re


class ParamMixin:
    @staticmethod
    def get_params():
        if request.method in ['GET', 'DELETE']:
            return to_lowcase({**request.args, **request.view_args})
        elif 'multipart/form-data' in request.content_type:
            return {
                **request.files,
                **to_lowcase(request.form.to_dict())
            }
        else:
            try:
                params = request.get_json()
                return iter_lowcase(params)
            except:
                return {
                    'data': request.get_data()
                }

    @staticmethod
    def clean_params(params):
        return {k: v for k, v in params.items() if v not in ('', 'null', None)}


def get_resource_classes_from_module(module):
    resource_classes = []
    for name, obj in inspect.getmembers(module):
        if inspect.isclass(obj) and name.endswith("Resource"):
            resource_classes.append(obj)
    return resource_classes


def get_resource_classes(resources_file):
    path = os.path.dirname(resources_file)
    resource_classes = []
    pattern = re.compile(r'.*resource(?!s).*\.py$')
    for root, dirs, files in os.walk(path):
        for file in files:
            if pattern.match(file):
                file_path = os.path.join(root, file)
                spec = importlib.util.spec_from_file_location('module', file_path)
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)
                resource_classes += get_resource_classes_from_module(module)
    return resource_classes


class Router(ParamMixin, Blueprint):
    """路由"""

    def __init__(self, *args, **kwargs):
        Blueprint.__init__(self, *args, **kwargs)

    def verify_user(self):
        """用户验证，通过继承覆盖此方法实现具体逻辑"""
        return True

    def handle_error(self, e):
        """错误处理，通过继承覆盖此方法实现具体逻辑"""
        pass

    @staticmethod
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)
        return wrapper

    def wrap_view_func(self, func, public=False):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            params = self.clean_params(self.get_params())
            if not self.verify_user() and not public:
                return {'code': 401, 'msg': '用户无权限'}
            try:
                data = self.decorator(func)(**params)
            except MessagePrompt as e:
                return {'code': 500, 'msg': str(e)}
            except ValidationError as e:
                return {'code': 400, 'msg': str(e)}
            except Exception as e:
                self.handle_error(e)
                raise e
            if isinstance(data, dict):
                data = iter_camel(data)
            elif isinstance(data, list):
                data = [iter_camel(item) if isinstance(item, dict) else item
                        for item in data]
            elif data is None:
                pass
            elif isinstance(data, str):
                pass 
            else:
                return data
            return {'code': 200, 'msg': '操作成功', 'data': data}
        return wrapper

    def add_resource(self, rule, resource_class):
        http_methods = ['get', 'post', 'put', 'delete']

        for method_name in http_methods:
            if method_name in dir(resource_class):
                method = getattr(resource_class, method_name)
                public = getattr(method, 'public', False)
                method = functools.partial(method)
                endpoint = str(random.randint(10000000, 99999999))
                self.add_url_rule(rule, endpoint, self.wrap_view_func(method,
                                                                      public=public), methods=[method_name.upper()])

    def add_resources(self, resource_classes):
        for resource_class in resource_classes:
            self.add_resource(resource_class.path, resource_class)

    def auto_add_resources(self, resources_file):
        resource_classes = get_resource_classes(resources_file)
        self.add_resources(resource_classes)

    def add(self, rule, public=False, **options):
        def decorator(f):
            endpoint = options.pop("endpoint", None)
            self.add_url_rule(rule, endpoint, self.wrap_view_func(
                f, public=public,), **options)
        return decorator
